@@ -1,14 +1,16 @@
-/* r2ghidra - LGPL - Copyright 2020-2024 - pancake, FXTi */
+/* r2ghidra - LGPL - Copyright 2020-2025 - pancake, FXTi */
+
+#include "SleighAsm.h"
+#include "SleighAnalValue.h"
+#include "ArchMap.h"
 
 #include <r_lib.h>
 #include <r_anal.h>
+
 #include <algorithm>
 #include <cfloat>
 #include <cmath>
 #include <cfenv>
-#include "ArchMap.h"
-#include "SleighAsm.h"
-#include "SleighAnalValue.h"
 
 #if R2_VERSION_NUMBER >= 50709
 #define RAnalEsil REsil
@@ -24,14 +26,15 @@
 
 // XXX dont use globals
 static SleighAsm *sanal = nullptr;
+#if R2_VERSION_NUMBER < 50909
 extern "C" RCore *Gcore;
+#endif
 
 static char *slid(RCore *core, const char *cpu, int bits, bool be) {
 	R_LOG_DEBUG ("slid (%s:%d:%d)", cpu, bits, be);
 	if (core == nullptr) {
-		core = Gcore;
-	} else if (Gcore == nullptr) {
-		Gcore = core;
+		R_LOG_ERROR ("cannot slid with nul core");
+		return NULL;
 	}
 	if (sanal == nullptr) {
 		sanal = new SleighAsm ();
@@ -72,7 +75,12 @@ static char *slid_arch(RAnal *anal) {
 #if R2_VERSION_NUMBER >= 50809
 
 extern "C" int archinfo(RArchSession *as, ut32 query) {
-	r_return_val_if_fail (as, 1);
+	R_RETURN_VAL_IF_FAIL (as, 1);
+#if R2_VERSION_NUMBER >= 50909
+	RBin *bin = (RBin*)as->arch->binb.bin;
+	RIO *io = (RIO*)bin->iob.io;
+	RCore *Gcore = (RCore *)io->coreb.core;
+#endif
 	char *arch = slid_arch (Gcore->anal); // is this initializing sanal global ptr?
 	if (sanal != nullptr) {
 		switch (query) {
@@ -97,7 +105,7 @@ extern "C" int archinfo(RArchSession *as, ut32 query) {
 #else
 extern "C" int archinfo(RAnal *anal, int query) {
 	// This is to check if RCore plugin set cpu properly.
-	r_return_val_if_fail (anal, -1);
+	R_RETURN_VAL_IF_FAIL (anal, -1);
 #if R2_VERSION_NUMBER >= 50609
 	if (R_STR_ISEMPTY (anal->config->cpu)) {
 		return -1;
@@ -1742,8 +1750,14 @@ extern "C" int sleigh_op(RAnal *a, RAnalOp *anal_op, ut64 addr, const ut8 *data,
 #if R2_VERSION_NUMBER >= 50809
 extern "C" bool sleigh_decode(RArchSession *as, RAnalOp *aop, RArchDecodeMask mask) {
 	REsil *esil = as->arch->esil;
+#if R2_VERSION_NUMBER >= 50909
+	RBin *bin = (RBin*)as->arch->binb.bin;
+	RIO *io = (RIO*)bin->iob.io;
+	RCore *Gcore = (RCore *)io->coreb.core;
+#else
 	RIO *io = Gcore->io;
 	RBin *bin = Gcore->bin;
+#endif
 	RAnal *anal = Gcore->anal;
 	if (bin != nullptr && esil != nullptr) {
 		io = bin->iob.io;
@@ -1950,8 +1964,12 @@ static std::string regtype_name(const char *cpu, const std::string &regname) {
 
 #if R2_VERSION_NUMBER >= 50809
 extern "C" char *r2ghidra_regs(RArchSession *as) {
-	r_return_val_if_fail (as, nullptr);
-
+	R_RETURN_VAL_IF_FAIL (as, nullptr);
+#if R2_VERSION_NUMBER >= 50909
+	RBin *bin = (RBin*)as->arch->binb.bin;
+	RIO *io = (RIO*)bin->iob.io;
+	RCore *Gcore = (RCore *)io->coreb.core;
+#endif
 	const char *cpu = r_config_get (Gcore->config, "asm.cpu"); // (as->config != nullptr)? as->config->cpu: "arm";
 
 	if (R_STR_ISEMPTY (cpu)) {
@@ -2018,7 +2036,7 @@ extern "C" char *r2ghidra_regs(RArchSession *as) {
 
 #else
 extern "C" char *get_reg_profile(RAnal *anal) {
-	r_return_val_if_fail (anal, nullptr);
+	R_RETURN_VAL_IF_FAIL (anal, nullptr);
 #if R2_VERSION_NUMBER >= 50609
 	const char *cpu = anal->config->cpu;
 #else
