@@ -1,12 +1,17 @@
-/* r2ghidra - LGPL - Copyright 2019-2023 - thestr4ng3r, pancake */
+/* r2ghidra - LGPL - Copyright 2019-2025 - thestr4ng3r, pancake */
 
 #include "R2Architecture.h"
 #include "CodeXMLParse.h"
 #include "PrettyXmlEncode.h"
-#include "ArchMap.h"
+#include "R2PrintC.h"
+
 #include "SleighAsm.h"
+#include "ArchMap.h"
+#include "PcodeFixupPreprocessor.h"
+
 #include "r2ghidra.h"
-#include "r_anal.h"
+
+#include <r_core.h>
 
 // Windows clash
 #ifdef restrict
@@ -14,18 +19,15 @@
 #endif
 
 #include <libdecomp.hh>
-#include <printc.hh>
-#include "R2PrintC.h"
-
-#include <r_core.h>
-#include <r_arch.h>
 
 #include <vector>
 #include <mutex>
 
 #undef DEBUG_EXCEPTIONS
 
+#if R2_VERSION_NUMBER < 50909
 extern "C" RCore *Gcore;
+#endif
 
 typedef bool (*ConfigVarCb)(void *user, void *data);
 
@@ -45,10 +47,10 @@ public:
 	const char *GetDesc() const { return desc; }
 	ConfigVarCb GetCallback() const	{ return callback; }
 
-	ut64 GetInt(RConfig *cfg) const	{ return r_config_get_i(cfg, name.c_str ()); }
-	bool GetBool(RConfig *cfg) const { return GetInt(cfg) != 0; }
-	std::string GetString(RConfig *cfg) const { return r_config_get(cfg, name.c_str ()); }
-	void Set(RConfig *cfg, const char *s) const { r_config_set(cfg, name.c_str (), s); }
+	ut64 GetInt(RConfig *cfg) const	{ return r_config_get_i (cfg, name.c_str ()); }
+	bool GetBool(RConfig *cfg) const { return GetInt (cfg) != 0; }
+	std::string GetString(RConfig *cfg) const { return r_config_get (cfg, name.c_str ()); }
+	void Set(RConfig *cfg, const char *s) const { r_config_set (cfg, name.c_str (), s); }
 	static const std::vector<const ConfigVar *> &GetAll() { return vars_all; }
 };
 
@@ -72,6 +74,7 @@ CV cfg_var_maximplref ("maximplref",  "2",        "Maximum number of references
 CV cfg_var_rawptr     ("rawptr",      "true",     "Show unknown globals as raw addresses instead of variables");
 CV cfg_var_verbose    ("verbose",     "false",    "Show verbose warning messages while decompiling");
 CV cfg_var_casts      ("casts",       "false",    "Show type casts where needed");
+CV cfg_var_fixups     ("fixups",      "false",    "Apply pcode fixups");
 CV cfg_var_roprop     ("roprop",      "0",        "Propagate read-only constants (0,1,2,3,4)");
 CV cfg_var_timeout    ("timeout",     "0",        "Run decompilation in a separate process and kill it after a specific time");
 
@@ -79,12 +82,20 @@ CV cfg_var_timeout    ("timeout",     "0",        "Run decompilation in a separa
 static std::recursive_mutex decompiler_mutex;
 
 class DecompilerLock {
+private:
+	RCore *_core;
 public:
-	DecompilerLock() {
+	DecompilerLock(RCore *core) : _core (core) {
 		if (!decompiler_mutex.try_lock()) {
+#if R2_VERSION_NUMBER >= 50909
+			void *bed = r_cons_sleep_begin (_core->cons);
+			decompiler_mutex.lock ();
+			r_cons_sleep_end (_core->cons, bed);
+#else
 			void *bed = r_cons_sleep_begin ();
 			decompiler_mutex.lock ();
 			r_cons_sleep_end (bed);
+#endif
 		}
 	}
 
@@ -111,7 +122,11 @@ static void PrintUsage(const RCore *const core) {
 		"%SLEIGHHOME" , "", "# Path to ghidra sleigh directory (same as r2ghidra.sleighhome)",
 		NULL
 	};
+#if R2_VERSION_NUMBER >= 50909
+	r_cons_cmd_help (core->cons, help, core->print->flags & R_PRINT_FLAGS_COLOR);
+#else
 	r_cons_cmd_help (help, core->print->flags & R_PRINT_FLAGS_COLOR);
+#endif
 }
 
 enum class DecompileMode {
@@ -169,6 +184,11 @@ static void Decompile(RCore *core, ut64 addr, DecompileMode mode, std::stringstr
 	}
 	arch.getCore()->sleepBegin ();
 	auto action = arch.allacts.getCurrent ();
+
+	if (cfg_var_fixups.GetBool (core->config)) {
+		PcodeFixupPreprocessor::fixupSharedReturnJumpToRelocs(function, func, core, arch);
+	}
+
 	int res;
 #ifndef DEBUG_EXCEPTIONS
 	try {
@@ -241,7 +261,7 @@ static void Decompile(RCore *core, ut64 addr, DecompileMode mode, std::stringstr
 }
 
 R_API RCodeMeta *r2ghidra_decompile_annotated_code(RCore *core, ut64 addr) {
-	DecompilerLock lock;
+	DecompilerLock lock(core);
 	RCodeMeta *code = nullptr;
 #ifndef DEBUG_EXCEPTIONS
 	try {
@@ -261,7 +281,7 @@ R_API RCodeMeta *r2ghidra_decompile_annotated_code(RCore *core, ut64 addr) {
 }
 
 static void DecompileCmd (RCore *core, DecompileMode mode) {
-	DecompilerLock lock;
+	DecompilerLock lock(core);
 
 #ifndef DEBUG_EXCEPTIONS
 	try {
@@ -269,45 +289,88 @@ static void DecompileCmd (RCore *core, DecompileMode mode) {
 		RCodeMeta *code = nullptr;
 		std::stringstream out_stream;
 #if R2_VERSION_NUMBER >= 50909
-		Decompile(core, core->addr, mode, out_stream, &code);
+		Decompile (core, core->addr, mode, out_stream, &code);
 #else
-		Decompile(core, core->offset, mode, out_stream, &code);
+		Decompile (core, core->offset, mode, out_stream, &code);
 #endif
 		switch (mode) {
 		case DecompileMode::DISASM:
 			{
-#if R2_VERSION_NUMBER >= 50403
 				RVector *offsets = r_codemeta_line_offsets (code);
-				r_codemeta_print_disasm (code, offsets, core->anal);
-				r_vector_free(offsets);
+#if R2_VERSION_NUMBER >= 50909
+				char *s = r_codemeta_print_disasm (code, offsets, core->anal);
+				r_cons_print (core->cons, s);
+				free (s);
 #else
-				RVector *offsets = r_codemeta_line_offsets (code);
-				r_codemeta_print (code, offsets);
-				r_vector_free (offsets);
+				r_codemeta_print_disasm (code, offsets, core->anal);
 #endif
+				r_vector_free (offsets);
 			}
 			break;
 		case DecompileMode::OFFSET:
 			{
 				RVector *offsets = r_codemeta_line_offsets (code);
+#if R2_VERSION_NUMBER >= 60003
+				char *s = r_codemeta_print2 (code, offsets, core->anal);
+				r_cons_print (core->cons, s);
+				free (s);
+#elif R2_VERSION_NUMBER >= 50909
+				char *s = r_codemeta_print (code, offsets);
+				r_cons_print (core->cons, s);
+				free (s);
+#else
 				r_codemeta_print (code, offsets);
+				r_cons_flush ();
+#endif
 				r_vector_free (offsets);
 			}
 			break;
 		case DecompileMode::DEFAULT:
-			r_codemeta_print (code, nullptr);
+			{
+#if R2_VERSION_NUMBER >= 60003
+				char *s = r_codemeta_print2 (code, nullptr, core->anal);
+				r_cons_print (core->cons, s);
+				free (s);
+#elif R2_VERSION_NUMBER >= 50909
+				char *s = r_codemeta_print (code, nullptr);
+				r_cons_print (core->cons, s);
+				free (s);
+#else
+				r_codemeta_print (code, nullptr);
+#endif
+			}
 			break;
 		case DecompileMode::STATEMENTS:
+#if R2_VERSION_NUMBER >= 50909
+			{
+				char *s = r_codemeta_print_comment_cmds (code);
+				r_cons_print (core->cons, s);
+				free (s);
+			}
+#else
 			r_codemeta_print_comment_cmds (code);
+#endif
 			break;
 		case DecompileMode::JSON:
-			r_codemeta_print_json (code);
+			{
+#if R2_VERSION_NUMBER >= 50909
+				char *s = r_codemeta_print_json (code);
+				r_cons_println (core->cons, s);
+				free (s);
+#else
+				r_codemeta_print_json (code);
+#endif
+			}
 			break;
 		case DecompileMode::XML:
 			out_stream << "</code></result>";
 			// fallthrough
 		default:
+#if R2_VERSION_NUMBER >= 50909
+			r_cons_printf (core->cons, "%s\n", out_stream.str().c_str ());
+#else
 			r_cons_printf ("%s\n", out_stream.str().c_str ());
+#endif
 			break;
 		}
 		r_codemeta_free (code);
@@ -323,7 +386,11 @@ static void DecompileCmd (RCore *core, DecompileMode mode) {
 				pj_s (pj, s.c_str ());
 				pj_end (pj);
 				pj_end (pj);
+#if R2_VERSION_NUMBER >= 50909
+				r_cons_printf (core->cons, "%s\n", pj_string (pj));
+#else
 				r_cons_printf ("%s\n", pj_string (pj));
+#endif
 				pj_free (pj);
 			}
 		} else {
@@ -340,7 +407,11 @@ public:
 		std::stringstream ss;
 		addr.printRaw (ss);
 		ss << ": " << mnem << ' ' << body;
+#if R2_VERSION_NUMBER >= 50909
+		r_cons_gprintf ("%s\n", ss.str().c_str ());
+#else
 		r_cons_printf ("%s\n", ss.str().c_str ());
+#endif
 	}
 };
 
@@ -414,7 +485,11 @@ public:
 				print_vardata (ss, vars[i]);
 			}
 		}
+#if R2_VERSION_NUMBER >= 50909
+		r_cons_gprintf ("    %s\n", ss.str().c_str ());
+#else
 		r_cons_printf ("    %s\n", ss.str().c_str ());
+#endif
 	}
 };
 
@@ -462,7 +537,7 @@ static void SetInitialSleighHome(RConfig *cfg) {
 }
 
 static void ListSleighLangs(RCore *core) {
-	DecompilerLock lock;
+	DecompilerLock lock(core);
 	R2Architecture::collectSpecFiles (std::cerr);
 	auto langs = R2Architecture::getLanguageDescriptions ();
 	if (langs.empty()) {
@@ -474,16 +549,24 @@ static void ListSleighLangs(RCore *core) {
 		return lang.getId ();
 	});
 	std::sort (ids.begin (), ids.end ());
-	std::for_each (ids.begin (), ids.end (), [](const std::string &id) {
+	std::for_each (ids.begin (), ids.end (), [core](const std::string &id) {
+#if R2_VERSION_NUMBER >= 50909
+		r_cons_printf (core->cons, "%s\n", id.c_str ());
+#else
 		r_cons_printf ("%s\n", id.c_str ());
+#endif
 	});
 }
 
 static void PrintAutoSleighLang(RCore *core) {
-	DecompilerLock lock;
+	DecompilerLock lock (core);
 	try {
 		auto id = SleighIdFromCore (core);
+#if R2_VERSION_NUMBER >= 50909
+		r_cons_printf (core->cons, "%s\n", id.c_str ());
+#else
 		r_cons_printf ("%s\n", id.c_str ());
+#endif
 	} catch (LowlevelError &e) {
 		R_LOG_WARN ("%s", e.explain.c_str ());
 	}
@@ -560,7 +643,11 @@ static void _cmd(RCore *core, const char *input) {
 		}
 		if (pid == 0) {
 			runcmd (core, input);
+#if R2_VERSION_NUMBER >= 50909
+			r_cons_flush (core->cons);
+#else
 			r_cons_flush ();
+#endif
 			fflush (stdout);
 			write (fds[1], "\x12", 1);
 			exit (0);
@@ -596,6 +683,28 @@ static void _cmd(RCore *core, const char *input) {
 	}
 }
 
+#if R2_VERSION_NUMBER >= 50909
+extern "C" bool r2ghidra_core_cmd(RCorePluginSession *cps, const char *input) {
+	RCore *core = cps->core;
+	if (!strcmp (input, "pdc.")) {
+		r_cons_println (core->cons, "pdc.ghidra");
+		return false;
+	}
+	if (r_str_startswith (input, "pdc.ghidra")) {
+		_cmd (core, input + strlen ("pdc.ghidra"));
+		return true;
+	}
+	if (r_str_startswith (input, "pdc.g")) {
+		_cmd (core, input + strlen ("pdc.g"));
+		return true;
+	}
+	if (r_str_startswith (input, "pdg")) {
+		_cmd (core, input + strlen ("pdg"));
+		return true;
+	}
+	return false;
+}
+#else
 extern "C" int r2ghidra_core_cmd(void *user, const char *input) {
 	RCore *core = (RCore *) user;
 	if (r_str_startswith (input, "pdg")) {
@@ -604,11 +713,12 @@ extern "C" int r2ghidra_core_cmd(void *user, const char *input) {
 	}
 	return false;
 }
+#endif
 
 bool SleighHomeConfig(void */* user */, void *data) {
 	std::lock_guard<std::recursive_mutex> lock(decompiler_mutex);
-	auto node = reinterpret_cast<RConfigNode *>(data);
-	SleighArchitecture::shutdown();
+	RConfigNode *node = reinterpret_cast<RConfigNode *>(data);
+	SleighArchitecture::shutdown ();
 	SleighArchitecture::specpaths = FileManage ();
 	if (R_STR_ISNOTEMPTY (node->value)) {
 		SleighArchitecture::scanForSleighDirectories (node->value);
@@ -622,11 +732,30 @@ extern "C" RArchPlugin r_arch_plugin_ghidra;
 extern "C" RAnalPlugin r_anal_plugin_ghidra;
 #endif
 
+#if R2_VERSION_NUMBER >= 50909
+extern "C" bool r2ghidra_core_init(RCorePluginSession *cps) {
+	std::lock_guard<std::recursive_mutex> lock(decompiler_mutex);
+	startDecompilerLibrary (nullptr);
+	RCore *core = reinterpret_cast<RCore *>(cps->core);
+	r_arch_plugin_add (core->anal->arch, &r_arch_plugin_ghidra);
+	RConfig *cfg = core->config;
+	r_config_lock (cfg, false);
+	for (const auto var : ConfigVar::GetAll ()) {
+		RConfigNode *node = var->GetCallback()
+			? r_config_set_cb (cfg, var->GetName (), var->GetDefault (), var->GetCallback ())
+			: r_config_set (cfg, var->GetName (), var->GetDefault ());
+		r_config_node_desc (node, var->GetDesc ());
+	}
+	r_config_lock (cfg, true);
+	SetInitialSleighHome (cfg);
+	return true;
+}
+#else
 extern "C" int r2ghidra_core_init(void *user, const char *cmd) {
 	std::lock_guard<std::recursive_mutex> lock(decompiler_mutex);
 	startDecompilerLibrary (nullptr);
-	auto *rcmd = reinterpret_cast<RCmd *>(user);
-	auto *core = reinterpret_cast<RCore *>(rcmd->data);
+	RCmd *rcmd = reinterpret_cast<RCmd *>(user);
+	RCore *core = reinterpret_cast<RCore *>(rcmd->data);
 	Gcore = core;
 #if R2_VERSION_NUMBER >= 50809
 	r_arch_plugin_add (core->anal->arch, &r_arch_plugin_ghidra);
@@ -645,9 +774,18 @@ extern "C" int r2ghidra_core_init(void *user, const char *cmd) {
 	SetInitialSleighHome (cfg);
 	return true;
 }
+#endif
 
+#if R2_VERSION_NUMBER >= 50909
+extern "C" bool r2ghidra_core_fini(RCorePluginSession *cps, const char *cmd) {
+	std::lock_guard<std::recursive_mutex> lock (decompiler_mutex);
+	shutdownDecompilerLibrary ();
+	return true;
+}
+#else
 extern "C" int r2ghidra_core_fini(void *user, const char *cmd) {
 	std::lock_guard<std::recursive_mutex> lock (decompiler_mutex);
 	shutdownDecompilerLibrary ();
 	return true;
 }
+#endif
